import numpy as np

from ._utils import _induce_cliques, _validate_geometry_input


def _build_coplanarity_node_lookup(geoms):
    """
    Identify coplanar points and create a look-up table for the coplanar geometries.
    Same function as in ``graph._utils``, but need to keep the index to use as graph ids
    """
    # geoms = geoms.reset_index(drop=True)
    coplanar = []
    nearest = []
    r = geoms.groupby(geoms).groups
    for g in r.values():
        if len(g) == 2:
            coplanar.append(g[0])
            nearest.append(g[1])
        elif len(g) > 2:
            for n in g[1:]:
                coplanar.append(n)
                nearest.append(g[0])
    return np.asarray(coplanar), np.asarray(nearest)


def pdna_to_adj(origins, network, node_ids, threshold):
    """Create an adjacency list of shortest network-based travel between
       origins and destinations in a pandana.Network.

    Parameters
    ----------
    origins : geopandas.GeoDataFrame
        Geodataframe of origin geometries to begin routing.
    network : pandana.Network
        pandana.Network instance that stores the local travel network
    node_ids:
        array of node_ids in the pandana.Network aligned with the input
        observations in ``origins``. This is created via a call like
        ``pandana.Network.get_node_ids(df.geometry.x, df.geometry.y)``
    threshold : int
        maximum travel distance (inclusive)

    Returns
    -------
    pandas.DataFrame
        adjacency list with columns 'origin', 'destination', and 'cost'
    """

    # map node ids in the network to index in the gdf
    mapper = dict(zip(node_ids, origins.index.values, strict=False))

    namer = {"source": "origin", network.impedance_names[0]: "cost"}

    adj = network.nodes_in_range(node_ids, threshold)
    adj = adj.rename(columns=namer)
    # swap osm ids for gdf index
    adj = adj.set_index("destination").rename(index=mapper).reset_index()
    adj = adj.set_index("origin").rename(index=mapper).reset_index()
    adj = adj[adj.destination.isin(origins.index.values)]

    return adj


def build_travel_graph(df, network, threshold, mapping_distance):
    """Compute the shortest path between gdf centroids via a pandana.Network
    and return an adjacency list with weight=cost. Note unlike distance_band,
    :math:`G_{ij}` and :math:`G_{ji}` are often different because travel networks
    may be directed.

    Parameters
    ----------
    df : geopandas.GeoDataFrame
        geodataframe of observations. CRS should be the same as the locations
        of ``node_x`` and ``node_y`` in the pandana.Network (usually 4326 if network
        comes from OSM, but sometimes projected to improve snapping quality).
    network : pandana.Network
        Network that encodes travel costs. See <https://udst.github.io/pandana/>
    threshold : int
        maximum travel cost to consider neighbors
    mapping_distance : int
        snapping tolerance passed to ``pandana.Network.get_node_ids`` that defines
        the maximum range at which observations are snapped to nearest nodes in the
        network. Default is None

    Returns
    -------
    pandas.Series
        adjacency formatted as multiindexed (focal, neighbor) series
    """
    df = df.copy()
    _validate_geometry_input(df.geometry, ids=None, valid_geometry_types="Point")
    df["node_ids"] = network.get_node_ids(
        df.geometry.x, df.geometry.y, mapping_distance
    )

    # depending on density of the graph nodes / observations, it is common to have
    # multiple observations snapped to the same network node, so use the clique
    # expansion logic to handle these cases

    # get indices of multi-observations at unique nodes
    coplanar, nearest = _build_coplanarity_node_lookup(df["node_ids"])
    # create adjacency on unique nodes
    adj = pdna_to_adj(df, network, df["node_ids"].values, threshold)
    # add clique members back to adjacency
    adj_cliques = _induce_cliques(
        adj.rename(
            columns={"origin": "focal", "destination": "neighbor", "cost": "weight"}
        ),
        coplanar=coplanar,
        nearest=nearest,
    )
    # reorder, drop induced dupes, and return
    adj_cliques = (
        adj_cliques.groupby(["focal", "neighbor"])
        .first()
        .reindex(df.index, level=0)
        .reindex(df.index, level=1)
        .reset_index()
    )

    return adj_cliques



#!/usr/bin/env python
"""
Build the documentation for the member comparison of W and Graph


"""

import inspect
import pandas as pd
import geopandas as gpd
import numpy as np
from libpysal.io import open as psopen
from libpysal import weights
from libpysal import graph
from libpysal import examples


examples.explain('sids2')


# Read the file in
gdf = gpd.read_file(examples.get_path('sids2.shp'))
gdf = gdf.set_crs('epsg:4326')

# Make weights and graph
w_queen = weights.Queen.from_dataframe(gdf)
g_queen = graph.Graph.build_contiguity(gdf, rook=False)


g_members = set(dir(g_queen))
w_members = set(dir(w_queen))


# filter out private members
g_members = {attr for attr in g_members if not attr.startswith('_')}
w_members = {attr for attr in w_members if not attr.startswith('_')}

compat = []
changed = []

for member in g_members & w_members:
    g_member = getattr(g_queen, member)
    w_member = getattr(w_queen, member)
    print(member, type(g_member), type(w_member), w_member.__class__.__name__)
    if type(g_member) == type(w_member):
        compat.append(member)
    else:
        changed.append(member)

changed.sort()
compat.sort()


changed_content = []
header = "Member, W Type, Graph Type"
changed_content.append(header)
for member in changed:
    line = [member]
    line.append(getattr(w_queen, member).__class__.__name__)
    line.append(getattr(g_queen, member).__class__.__name__)
    changed_content.append(",".join(line))

changed_content = [line.split(",") for line in changed_content]


def create_rst_table(data):
    if not data or not all(isinstance(row, list) for row in data):
        raise ValueError("Input should be a list of lists")

    # Determine the width of each column
    col_widths = [max(len(str(item)) for item in column)
                  for column in zip(*data)]

    # Function to create a row separator
    def create_separator(char):
        return "+" + "+".join(char * (width + 2) for width in col_widths) + "+"

    # Function to create a row
    def create_row(row):
        return "|" + "|".join(f" {str(item).ljust(width)} " for item, width in zip(row, col_widths)) + "|"

    # Create the table
    table = []
    table.append(create_separator('-'))
    table.append(create_row(data[0]))
    table.append(create_separator('='))
    for row in data[1:]:
        table.append(create_row(row))
        table.append(create_separator('-'))

    return "\n".join(table)


changed_table = create_rst_table(changed_content)


content = """
W to Graph Member Comparisions
==============================


Overview
--------

This guide compares the members (attributes and methods) from the
`W` class and the `Graph` class.

It is intended for developers. Users interested in migrating to the
new Graph class from W should see the `migration guide <user-guide/graph/w_g_migration.html>`_.


Members common to W and Graph
-----------------------------
"""


common_content = []
header = "Member,  Typee"
common_content.append(header)
for member in compat:

    line = [
        f"`{member} <generated/libpysal.graph.Graph.html#libpysal.graph.Graph.{member}>`_"]
    label = ":attr"
    ga = getattr(g_queen, member)
    class_type = type(ga)
    gat = f"{class_type.__module__}.{class_type.__name__}"
    if inspect.ismethod(ga):
        label = ":meth"
    gs = f" {gat}"

    line.append(gs)

    common_content.append(",".join(line))

common_content = [line.split(",") for line in common_content]


content = f"{content}\n\n{create_rst_table(common_content)}"
head = """
Members common to W and Graph with different types
--------------------------------------------------

"""

changed_content = []
header = "Member, Queen Type, Graph Type"
changed_content.append(header)
for member in changed:
    ms = f"{member}"
    line = [ms]
    ga = getattr(g_queen, member)
    class_type = type(ga)
    gat = f"{class_type.__module__}.{class_type.__name__}"
    gat = f"`{gat} <generated/libpysal.graph.Graph.html#libpysal.graph.Graph.{member}>`_"

    gs = f"{gat}"

    wa = getattr(w_queen, member)
    class_type = type(wa)
    wat = f"{class_type.__module__}.{class_type.__name__}"
    wat = f"`{wat} <generated/libpysal.weights.W.html#libpysal.weights.W.{member}>`_"

    ws = f"{wat}"
    line.append(ws)
    line.append(gs)

    changed_content.append(",".join(line))

changed_content = [line.split(",") for line in changed_content]
content = f"{content}\n\n{head}\n\n{create_rst_table(changed_content)}"


head = """
Members unique to W
-------------------

"""

content = f"{content}\n\n{head}"


w_only = [member for member in w_members - g_members]
w_only.sort()

w_content = []
header = "Member,  Type"
w_content.append(header)
for member in w_only:
    line = [
        f"`{member} <generated/libpysal.weights.W.html#libpysal.weights.W.{member}>`_"]
    wa = getattr(w_queen, member)
    class_type = type(wa)
    wat = f"{class_type.__module__}.{class_type.__name__}"

    ws = f"{wat}"

    line.append(ws)

    w_content.append(",".join(line))

w_content = [line.split(",") for line in w_content]
content = f"{content}\n\n{create_rst_table(w_content)}"


head = """
Members unique to Graph
-----------------------

"""

content = f"{content}\n\n{head}"


g_only = [member for member in g_members - w_members]
g_only.sort()

g_content = []
header = "Member,  Type"
g_content.append(header)
for member in g_only:
    line = [
        f"`{member} <generated/libpysal.graph.Graph.html#libpysal.graph.Graph.{member}>`_"]
    ga = getattr(g_queen, member)
    class_type = type(ga)
    gat = f"{class_type.__module__}.{class_type.__name__}"

    gs = f"{gat}"

    line.append(gs)

    g_content.append(",".join(line))

g_content = [line.split(",") for line in g_content]
content = f"{content}\n\n{create_rst_table(g_content)}"


with open("../docs/migration.rst", 'w') as guide:
    guide.write(content)



import os.path
from warnings import warn

from ...weights import W
from ...weights.util import remap_ids
from .. import fileio
from . import gwt

__author__ = "Myunghwa Hwang <mhwang4@gmail.com>"
__all__ = ["ArcGISTextIO"]


class ArcGISTextIO(gwt.GwtIO):
    """Opens, reads, and writes weights file objects in ArcGIS ASCII text format.
    Spatial weights objects in the ArcGIS text format are used in ArcGIS Spatial
    Statistics tools. This format is a simple text file with ASCII encoding and
    can be directly used with the tools under the category of "Mapping Clusters."
    But, it cannot be used with the "Generate Spatial Weights Matrix" tool.

    The first line of the ArcGIS text file is a header including the name of
    a data column that holded the ID variable in the original source data table.
    After this header line, it includes three data columns for origin ID,
    destination ID, and weight values. ArcGIS Spatial Statistics tools support
    only unique integer IDs. Thus, the values in the first two columns should
    be integers. For the case where a weights object uses non-integer IDs,
    `ArcGISTextIO` allows users to use internal IDs corresponding to record
    numbers, instead of original IDs.

    An exemplary structure of an ArcGIS text file is as follows:

    [Line 1]    StationID
    [Line 2]    1    1    0.0
    [Line 3]    1    2    0.1
    [Line 4]    1    3    0.14286
    [Line 5]    2    1    0.1
    [Line 6]    2    3    0.05
    [Line 7]    3    1    0.16667
    [Line 8]    3    2    0.06667
    [Line 9]    3    3    0.0


    As shown in the above example, this file format allows explicit specification
    of weights for self-neighbors. When no entry is available for self-neighbors,
    ArcGIS spatial statistics tools consider they have zero weights. The PySAL
    `ArcGISTextIO` class ignores self-neighbors if their weights are zero.

    References
    ----------
    http://webhelp.esri.com/arcgisdesktop/9.3/index.cfm?TopicName=Modeling_spatial_relationships

    Notes
    -----

    When there is a ``.dbf`` file whose name is identical to the name of the source
    text file, `ArcGISTextIO` checks the data type of the ID data column and uses it
    for reading and writing the text file. Otherwise, it considers IDs are strings.

    """

    FORMATS = ["arcgis_text"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        args = args[:2]
        gwt.GwtIO.__init__(self, *args, **kwargs)

    def _read(self):
        """Read in an ArcGIS text file.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        StopIteration
            Raised at the EOF.

        TypeError
            Raised when the IDs are not integers.

        Examples
        --------

        Type ``dir(w)`` at the interpreter to see what methods are supported.
        Open a text file and read it into a PySAL weights object.

        >>> import libpysal
        >>> w = libpysal.io.open(
        ...     libpysal.examples.get_path('arcgis_txt.txt'), 'r', 'arcgis_text'
        ... ).read()

        Get the number of observations from the header.

        >>> w.n
        3

        Get the mean number of neighbors.

        >>> w.mean_neighbors
        2.0

        Get neighbor distances for a single observation.

        >>> w[1]
        {2: 0.1, 3: 0.14286}

        """

        if self.pos > 0:
            raise StopIteration

        id_var = self.file.readline().strip()
        self.varName = id_var
        id_order = None
        id_type = int

        try:
            dbf = os.path.join(self.dataPath + ".dbf")
            if os.path.exists(dbf):
                db = fileio.FileIO(dbf, "r")
                if id_var in db.header:
                    id_order = db.by_col(id_var)
                    id_type = type(id_order[0])
                else:
                    msg = "ID_VAR:'%s' was in in the DBF header, "
                    msg += "proceeding with unordered string IDs."
                    msg = msg % id_var
                    warn(msg, RuntimeWarning, stacklevel=2)
            else:
                msg = "DBF relating to ArcGIS TEXT was not found, "
                msg += "proceeding with unordered string IDs."
                warn(msg, RuntimeWarning, stacklevel=2)
        except:  # noqa: E722
            msg = "Exception occurred will reading DBF, "
            msg += "proceeding with unordered string IDs."
            warn(msg, RuntimeWarning, stacklevel=2)

        if (id_type is not int) or (id_order and type(id_order)[0] is not int):
            raise TypeError("The data type for IDs should be integer.")

        if id_order:
            self.n = len(id_order)
            self.shp = os.path.split(self.dataPath)[1].split(".")[0]
        self.id_var = id_var

        weights, neighbors = self._readlines(id_type)
        for k in neighbors:
            if k in neighbors[k]:
                k_index = neighbors[k].index(k)
                if weights[k][k_index] == 0.0:
                    del neighbors[k][k_index]
                    del weights[k][k_index]

        self.pos += 1

        w = W(neighbors, weights)

        return w

    def write(self, obj, useIdIndex=False):  # noqa: N803
        """

        Parameters
        ----------
        obj : libpysal.weights.W
            A PySAL `W` object.
        useIdIndex : bool
            Use the `W` IDs and remap (``True``). Default is ``False``.

        Raises
        ------
        TypeError
            Raised when the IDs in input ``obj`` are not integers.
        TypeError
            Raised when the input ``obj`` is not a PySAL `W`.

        Examples
        --------

        >>> import tempfile, libpysal, os
        >>> testfile = libpysal.io.open(
        ...     libpysal.examples.get_path('arcgis_txt.txt'), 'r', 'arcgis_text'
        ... )
        >>> w = testfile.read()

        Create a temporary file for this example.

        >>> f = tempfile.NamedTemporaryFile(suffix='.txt')

        Reassign to a new variable.

        >>> fname = f.name

        Close the temporary named file.

        >>> f.close()

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w', 'arcgis_text')

        Write the Weights object into the open file.

        >>> o.write(w)
        >>> o.close()

        Read in the newly created text file.

        >>> wnew =  libpysal.io.open(fname, 'r', 'arcgis_text').read()

        Compare values from old to new.

        >>> wnew.pct_nonzero == w.pct_nonzero
        True

        Clean up the temporary file created for this example.

        >>> os.remove(fname)

        """

        self._complain_ifclosed(self.closed)

        if issubclass(type(obj), W):
            id_type = type(obj.id_order[0])

            if id_type is not int and not useIdIndex:
                raise TypeError("ArcGIS TEXT weight files support only integer IDs.")

            if useIdIndex:
                id2i = obj.id2i
                obj = remap_ids(obj, id2i)

            header = f"{self.varName}\n"
            self.file.write(header)
            self._writelines(obj)
        else:
            raise TypeError(f"Expected a PySAL weights object, got: {type(obj)}.")



import warnings

import numpy
from sklearn.metrics import pairwise_distances

from ._utils import _validate_geometry_input

_VALID_GEOMETRY_TYPES = ["Point"]


def _spatial_matching(
    x,
    y=None,
    n_matches=5,
    metric="euclidean",
    solver=None,
    return_mip=False,
    allow_partial_match=False,
    **metric_kwargs,
):
    """
    Match locations in one dataset to at least `n_matches`
    locations in another (possibly identical) dataset
    by minimizing the total distance between matched locations.

    Letting :math:`d_{ij}` be

    .. math::

        \\text{minimize} \\sum_i^n \\sum_j^n  d_{ij}m_{ij}

        \\text{subject to}
            \\sum_j^n m_{ij} >= k \\forall i

            m_{ij} \\in {0,1} \\forall ij


    Parameters
    ----------
    x : numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame
        geometries that need matches. If a geopandas.Geo* object
        is provided, the .geometry attribute is used. If a numpy.ndarray with
        a geometry dtype is used, then the coordinates are extracted and used.
    y : numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame (default: None)
        geometries that are used as a source for matching. If a geopandas.Geo* object
        is provided, the .geometry attribute is used. If a numpy.ndarray with
        a geometry dtype is used, then the coordinates are extracted and
        used. If none, matches are made within `x`.
    n_matches : int (default: None)
        number of matches
    metric : string or callable (default: 'euclidean')
        distance function to apply over the input coordinates. Supported options
        depend on whether or not scikit-learn is installed. If so, then any
        distance function supported by scikit-learn is supported here. Otherwise,
        only euclidean, minkowski, and manhattan/cityblock distances are admitted.
    solver : solver from pulp (default: None)
        a solver defined by the pulp optimization library. If no solver is
        provided, pulp's default solver will be used. This is generally
        pulp.COIN(), but this may vary depending on your configuration.
    return_mip : bool (default: False)
        whether or not to return the instance of the pulp.LpProblem. By
        default, the problem is not returned to the user.
    allow_partial_match : bool (default: False)
        whether to allow for partial matching. A partial match may have
        a weight between zero and one, while a "full" match (by default)
        must have a weight of either zero or one. A partial matching may
        have a shorter total distance, but will result in a weighted
        graph.
    """
    try:
        import pulp
    except ImportError as error:
        raise ImportError("spatial matching requires the pulp library") from error
    if metric == "precomputed":
        distance_matrix = x
        match_between = y is not None
    elif y is not None:
        x, x_ids, _ = _validate_geometry_input(
            x, ids=None, valid_geometry_types=_VALID_GEOMETRY_TYPES
        )
        y, y_ids, _ = _validate_geometry_input(
            y, ids=None, valid_geometry_types=_VALID_GEOMETRY_TYPES
        )
        distance_matrix = pairwise_distances(x, y, metric=metric)
        match_between = True
    else:
        x, x_ids, _ = _validate_geometry_input(
            x, ids=None, valid_geometry_types=_VALID_GEOMETRY_TYPES
        )
        y_ids = x_ids
        distance_matrix = pairwise_distances(x, metric=metric, **metric_kwargs)

        match_between = False

    n_targets, n_sources = distance_matrix.shape

    if match_between:
        row, col = numpy.meshgrid(
            numpy.arange(n_targets), numpy.arange(n_sources), indexing="ij"
        )
        row = row.flatten()
        col = col.flatten()
    else:
        # if we are matching within, we need to
        row, col = numpy.triu_indices(
            n=n_targets, m=n_sources, k=int(not match_between)
        )

    mp = pulp.LpProblem("optimal-neargraph", sense=pulp.LpMinimize)
    # a match is as binary decision variable, connecting observation i to observation j
    match_vars = pulp.LpVariable.dicts(
        "match",
        lowBound=0,
        upBound=1,
        indices=zip(row, col, strict=True),
        cat="Continuous" if allow_partial_match else "Binary",
    )
    # we want to minimize the geographic distance of links in the graph
    mp.objective = pulp.lpSum(
        [
            match_vars[i, j] * distance_matrix[i, j]
            for i, j in zip(row, col, strict=True)
        ]
    )

    # for each observation
    for j in range(n_targets):
        # there must be exactly k other matched observations, which might live
        if match_between:
            summand = pulp.lpSum(
                [
                    # over the whole match matrix
                    match_vars[j, i]
                    for i in range(n_sources)
                ]
            )
            sense = 1
        else:
            summand = pulp.lpSum(
                [
                    # in the "upper" triangle, or "lower" triangle
                    match_vars.get((i, j), match_vars.get((j, i)))
                    for i in range(n_sources)
                    if (i != j)
                ]
            )
            sense = int(not allow_partial_match)
        mp += pulp.LpConstraint(summand, sense=sense, rhs=n_matches)
    if match_between:  # but, we may choose to ignore some sources
        for i in range(n_sources):
            summand = pulp.lpSum([match_vars[j, i] for j in range(n_targets)])
            mp += pulp.LpConstraint(summand, sense=-1, rhs=n_matches)

    status = mp.solve(solver)

    if (status != 1) & (not allow_partial_match):
        warnings.warn(
            f"Problem is {pulp.LpStatus[status]}, so edge weights may be non-integer!",
            stacklevel=2,
        )

    edges = [
        (*key, value.value()) for key, value in match_vars.items() if value.value() > 0
    ]
    if not match_between:
        edges.extend([(*tuple(reversed(edge[:-1])), edge[-1]) for edge in edges])

    heads, tails, weights = map(numpy.asarray, zip(*sorted(edges), strict=True))

    if return_mip:
        return x_ids[heads], y_ids[tails], weights, mp
    return x_ids[heads], y_ids[tails], weights



"""
Weights.
"""

# ruff: noqa: N802, N803

__author__ = "Sergio J. Rey <srey@asu.edu>"

import copy
import math
import warnings
from collections import defaultdict
from os.path import basename

import numpy as np
import scipy.sparse
from scipy.sparse.csgraph import connected_components

from ..io.fileio import FileIO

# from .util import full, WSP2W resolve import cycle by
# forcing these into methods
from . import adjtools

__all__ = ["W", "WSP"]


class _LabelEncoder:
    """Encode labels with values between 0 and n_classes-1.

    Attributes
    ----------
    classes_: array of shape [n_classes]
        Class labels for each index.

    Examples
    --------
    >>> le = _LabelEncoder()
    >>> le.fit(["NY", "CA", "NY", "CA", "TX", "TX"])
    >>> le.classes_
    array(['CA', 'NY', 'TX'])
    >>> le.transform(["NY", "CA", "NY", "CA", "TX", "TX"])
    array([1, 0, 1, 0, 2, 2])
    """

    def fit(self, y):
        """Fit label encoder.

        Parameters
        ----------
        y : list
            list of labels

        Returns
        -------
        self : instance of self.
          Fitted label encoder.
        """
        self.classes_ = np.unique(y)
        return self

    def transform(self, y):
        """Transform labels to normalized encoding.

        Parameters
        ----------
        y : list
            list of labels

        Returns
        -------
        y : array
            array of normalized labels.
        """
        return np.searchsorted(self.classes_, y)


class W:
    """
    Spatial weights class. Class attributes are described by their
    docstrings. to view, use the ``help`` function.

    Parameters
    ----------
    neighbors : dict
        Key is region ID, value is a list of neighbor IDS.
        For example, ``{'a':['b'],'b':['a','c'],'c':['b']}``.
    weights : dict
       Key is region ID, value is a list of edge weights.
       If not supplied all edge weights are assumed to have a weight of 1.
       For example, ``{'a':[0.5],'b':[0.5,1.5],'c':[1.5]}``.
    id_order : list
       An ordered list of ids, defines the order of observations when
       iterating over ``W`` if not set, lexicographical ordering is used
       to iterate and the ``id_order_set`` property will return ``False``.
       This can be set after creation by setting the ``id_order`` property.
    silence_warnings : bool
       By default ``libpysal`` will print a warning if the dataset contains
       any disconnected components or islands. To silence this warning set this
       parameter to ``True``.
    ids : list
        Values to use for keys of the neighbors and weights ``dict`` objects.

    Attributes
    ----------
    asymmetries
    cardinalities
    component_labels
    diagW2
    diagWtW
    diagWtW_WW
    histogram
    id2i
    id_order
    id_order_set
    islands
    max_neighbors
    mean_neighbors
    min_neighbors
    n
    n_components
    neighbor_offsets
    nonzero
    pct_nonzero
    s0
    s1
    s2
    s2array
    sd
    sparse
    trcW2
    trcWtW
    trcWtW_WW
    transform

    Examples
    --------
    >>> from libpysal.weights import W
    >>> neighbors = {0: [3, 1], 1: [0, 4, 2], 2: [1, 5], 3: [0, 6, 4], 4: [1, 3, 7, 5], 5: [2, 4, 8], 6: [3, 7], 7: [4, 6, 8], 8: [5, 7]}
    >>> weights = {0: [1, 1], 1: [1, 1, 1], 2: [1, 1], 3: [1, 1, 1], 4: [1, 1, 1, 1], 5: [1, 1, 1], 6: [1, 1], 7: [1, 1, 1], 8: [1, 1]}
    >>> w = W(neighbors, weights)
    >>> "%.3f"%w.pct_nonzero
    '29.630'

    Read from external `.gal file <https://geodacenter.github.io/workbook/4a_contig_weights/lab4a.html#gal-weights-file>`_.

    >>> import libpysal
    >>> w = libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
    >>> w.n
    78
    >>> "%.3f"%w.pct_nonzero
    '6.542'

    Set weights implicitly.

    >>> neighbors = {0: [3, 1], 1: [0, 4, 2], 2: [1, 5], 3: [0, 6, 4], 4: [1, 3, 7, 5], 5: [2, 4, 8], 6: [3, 7], 7: [4, 6, 8], 8: [5, 7]}
    >>> w = W(neighbors)
    >>> round(w.pct_nonzero,3)
    29.63
    >>> from libpysal.weights import lat2W
    >>> w = lat2W(100, 100)
    >>> w.trcW2
    39600.0
    >>> w.trcWtW
    39600.0
    >>> w.transform='r'
    >>> round(w.trcW2, 3)
    2530.722
    >>> round(w.trcWtW, 3)
    2533.667

    Cardinality Histogram:

    >>> w.histogram
    [(2, 4), (3, 392), (4, 9604)]

    Disconnected observations (islands):

    >>> from libpysal.weights import W
    >>> w = W({1:[0],0:[1],2:[], 3:[]})

    UserWarning: The weights matrix is not fully connected:
    There are 3 disconnected components.
    There are 2 islands with ids: 2, 3.
    """  # noqa: E501

    def __init__(
        self,
        neighbors,
        weights=None,
        id_order=None,
        silence_warnings=False,
        ids=None,  # noqa: ARG002
    ):
        self.silence_warnings = silence_warnings
        self.transformations = {}
        self.neighbors = neighbors
        if not weights:
            weights = {}
            for key in neighbors:
                weights[key] = [1.0] * len(neighbors[key])
        self.weights = weights
        self.transformations["O"] = self.weights.copy()  # original weights
        self.transform = "O"
        if id_order is None:
            self._id_order = list(self.neighbors.keys())
            self._id_order.sort()
            self._id_order_set = False
        else:
            self._id_order = id_order
            self._id_order_set = True
        self._reset()
        self._n = len(self.weights)
        if (not self.silence_warnings) and (self.n_components > 1):
            message = (
                "The weights matrix is not fully connected: "
                "\n There are %d disconnected components." % self.n_components
            )
            ni = len(self.islands)
            if ni == 1:
                message = (
                    message + f"\n There is 1 island with id: {str(self.islands[0])}."
                )
            elif ni > 1:
                message = message + "\n There are %d islands with ids: %s." % (
                    ni,
                    ", ".join(str(island) for island in self.islands),
                )
            warnings.warn(message, stacklevel=2)

    def _reset(self):
        """Reset properties."""
        self._cache = {}

    def to_file(self, path="", format=None):  # noqa: A002
        """
        Write a weights to a file. The format is guessed automatically
        from the path, but can be overridden with the format argument.

        See libpysal.io.FileIO for more information.

        Parameters
        ----------
        path    :   string
                    location to save the file
        format  :   string
                    string denoting the format to write the weights to.

        Returns
        -------
        None
        """
        f = FileIO(dataPath=path, mode="w", dataFormat=format)
        f.write(self)
        f.close()

    @classmethod
    def from_file(cls, path="", format=None):  # noqa: A002
        """
        Read a weights file into a W object.

        Parameters
        ----------
        path    :   string
                    location to save the file
        format  :   string
                    string denoting the format to write the weights to.

        Returns
        -------
        W object
        """
        f = FileIO(dataPath=path, mode="r", dataFormat=format)
        w = f.read()
        f.close()
        return w

    @classmethod
    def from_shapefile(cls, *args, **kwargs):
        # we could also just "do the right thing," but I think it'd make sense to
        # try and get people to use `Rook.from_shapefile(shapefile)` rather than
        # W.from_shapefile(shapefile, type=`rook`), otherwise we'd need to build
        # a type dispatch table. Generic W should be for stuff we don't know
        # anything about.
        raise NotImplementedError(
            "Use type-specific constructors, like Rook, Queen, DistanceBand, or Kernel"
        )

    @classmethod
    def from_WSP(cls, WSP, silence_warnings=True):
        """Create a pysal W from a pysal WSP object (thin weights matrix).

        Parameters
        ----------
        wsp                     : WSP
                                PySAL sparse weights object
        silence_warnings        : bool
           By default ``libpysal`` will print a warning if the dataset contains
           any disconnected components or islands. To silence this warning set this
           parameter to ``True``.

        Returns
        -------
        w       : W
                PySAL weights object

        Examples
        --------
        >>> from libpysal.weights import lat2W, WSP, W

        Build a 10x10 scipy.sparse matrix for a rectangular 2x5 region of cells
        (rook contiguity), then construct a PySAL sparse weights object (wsp).

        >>> sp = lat2SW(2, 5)
        >>> wsp = WSP(sp)
        >>> wsp.n
        10
        >>> wsp.sparse[0].todense()
        matrix([[0, 1, 0, 0, 0, 1, 0, 0, 0, 0]], dtype=int8)

        Create a standard PySAL W from this sparse weights object.

        >>> w = W.from_WSP(wsp)
        >>> w.n
        10
        >>> print(w.full()[0][0])
        [0 1 0 0 0 1 0 0 0 0]
        """
        data = WSP.sparse.data
        indptr = WSP.sparse.indptr
        id_order = WSP.id_order
        if id_order:
            # replace indices with user IDs
            indices = [id_order[i] for i in WSP.sparse.indices]
        else:
            id_order = list(range(WSP.n))
        neighbors, weights = {}, {}
        start = indptr[0]
        for i in range(WSP.n):
            oid = id_order[i]
            end = indptr[i + 1]
            neighbors[oid] = indices[start:end]
            weights[oid] = data[start:end]
            start = end
        ids = copy.copy(WSP.id_order)
        w = W(neighbors, weights, ids, silence_warnings=silence_warnings)
        w._sparse = copy.deepcopy(WSP.sparse)
        w._cache["sparse"] = w._sparse
        return w

    @classmethod
    def from_adjlist(
        cls, adjlist, focal_col="focal", neighbor_col="neighbor", weight_col=None
    ):
        """
        Return an adjacency list representation of a weights object.

        Parameters
        ----------
        adjlist : pandas.DataFrame
            Adjacency list with a minimum of two columns.
        focal_col : str
            Name of the column with the "source" node ids.
        neighbor_col : str
            Name of the column with the "destination" node ids.
        weight_col : str
            Name of the column with the weight information. If not provided and
            the dataframe has no column named "weight" then all weights
            are assumed to be 1.
        """
        if weight_col is None:
            weight_col = "weight"
        try_weightcol = getattr(adjlist, weight_col)
        if try_weightcol is None:
            adjlist = adjlist.copy(deep=True)
            adjlist["weight"] = 1
        grouper = adjlist.groupby(focal_col)
        neighbors = {}
        weights = {}
        for ix, chunk in grouper:
            neighbors_to_ix = chunk[neighbor_col].values
            weights_to_ix = chunk[weight_col].values
            mask = neighbors_to_ix != ix
            neighbors[ix] = neighbors_to_ix[mask].tolist()
            weights[ix] = weights_to_ix[mask].tolist()
        return cls(neighbors=neighbors, weights=weights)

    def to_adjlist(
        self,
        remove_symmetric=False,
        drop_islands=None,
        focal_col="focal",
        neighbor_col="neighbor",
        weight_col="weight",
        sort_joins=False,
    ):
        """
        Compute an adjacency list representation of a weights object.

        Parameters
        ----------
        remove_symmetric : bool
            Whether or not to remove symmetric entries. If the ``W``
            is symmetric, a standard directed adjacency list will contain
            both the forward and backward links by default because adjacency
            lists are a directed graph representation. If this is ``True``,
            a ``W`` created from this adjacency list **MAY NOT BE THE SAME**
            as the original ``W``. If you would like to consider (1,2) and
            (2,1) as distinct links, leave this as ``False``.
        drop_islands : bool
            Whether or not to preserve islands as entries in the adjacency
            list. By default, observations with no neighbors do not appear
            in the adjacency list. If islands are kept, they are coded as
            self-neighbors with zero weight.
        focal_col : str
            Name of the column in which to store "source" node ids.
        neighbor_col : str
            Name of the column in which to store "destination" node ids.
        weight_col : str
            Name of the column in which to store weight information.
        sort_joins : bool
            Whether or not to lexicographically sort the adjacency
            list by (focal_col, neighbor_col). Default is False.
        """
        try:
            import pandas
        except (ImportError, ModuleNotFoundError):
            raise ImportError(
                "pandas must be installed & importable to use this method"
            ) from None
        if (drop_islands is None) and not (self.silence_warnings):
            warnings.warn(
                (
                    "In the next version of libpysal, observations with no neighbors "
                    "will be included in adjacency lists as loops (row with the same "
                    "focal and neighbor) with zero weight. In the current version, "
                    "observations with no neighbors are dropped. If you would like to "
                    "keep the current behavior, use drop_islands=True in this function"
                ),
                DeprecationWarning,
                stacklevel=2,
            )
            drop_islands = True

        focal_ix, neighbor_ix = self.sparse.nonzero()
        idxs = np.array(self.id_order)
        focal_ix = idxs[focal_ix]
        neighbor_ix = idxs[neighbor_ix]
        weights = self.sparse.data
        adjlist = pandas.DataFrame(
            {focal_col: focal_ix, neighbor_col: neighbor_ix, weight_col: weights}
        )
        if remove_symmetric:
            adjlist = adjtools.filter_adjlist(adjlist)
        if not drop_islands:
            island_adjlist = pandas.DataFrame(
                {focal_col: self.islands, neighbor_col: self.islands, weight_col: 0}
            )
            adjlist = pandas.concat((adjlist, island_adjlist)).reset_index(drop=True)
        if sort_joins:
            return adjlist.sort_values([focal_col, neighbor_col])
        return adjlist

    def to_networkx(self):
        """Convert a weights object to a ``networkx`` graph.

        Returns
        -------
        A ``networkx`` graph representation of the ``W`` object.
        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError(
                "NetworkX 2.7+ is required to use this function."
            ) from None
        g = nx.DiGraph() if len(self.asymmetries) > 0 else nx.Graph()
        return nx.from_scipy_sparse_array(self.sparse, create_using=g)

    @classmethod
    def from_networkx(cls, graph, weight_col="weight"):  # noqa: ARG003
        """Convert a ``networkx`` graph to a PySAL ``W`` object.

        Parameters
        ----------
        graph : networkx.Graph
            The graph to convert to a ``W``.
        weight_col : string
            If the graph is labeled, this should be the name of the field
            to use as the weight for the ``W``.

        Returns
        -------
        w : libpysal.weights.W
            A ``W`` object containing the same graph as the ``networkx`` graph.
        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError(
                "NetworkX 2.7+ is required to use this function."
            ) from None
        sparse_array = nx.to_scipy_sparse_array(graph)
        w = WSP(sparse_array).to_W()
        return w

    @property
    def sparse(self):
        """Sparse matrix object. For any matrix manipulations required for w,
        ``w.sparse`` should be used. This is based on ``scipy.sparse``.
        """
        if "sparse" not in self._cache:
            self._sparse = self._build_sparse()
            self._cache["sparse"] = self._sparse
        return self._sparse

    @classmethod
    def from_sparse(cls, sparse):
        """Convert a ``scipy.sparse`` array to a PySAL ``W`` object.

        Parameters
        ----------
        sparse : scipy.sparse array

        Returns
        -------
        w : libpysal.weights.W
            A ``W`` object containing the same graph as the ``scipy.sparse`` graph.

        Notes
        -----
        When the sparse array has a zero in its data attribute, and
        the corresponding row and column values are equal, the value
        for the pysal weight will be 0 for the "loop".
        """
        coo = sparse.tocoo()
        neighbors = defaultdict(list)
        weights = defaultdict(list)
        for k, v, w in zip(coo.row, coo.col, coo.data, strict=True):
            neighbors[k].append(v)
            weights[k].append(w)
        return W(neighbors=neighbors, weights=weights)

    def to_sparse(self, fmt="coo"):
        """Generate a ``scipy.sparse`` array object from a pysal W.

        Parameters
        ----------
        fmt : {'bsr', 'coo', 'csc', 'csr'}
          scipy.sparse format

        Returns
        -------
        scipy.sparse array
          A scipy.sparse array with a format of fmt.

        Notes
        -----
        The keys of the w.neighbors are encoded
        to determine row,col in the sparse array.
        """
        disp = {}
        disp["bsr"] = scipy.sparse.bsr_array
        disp["coo"] = scipy.sparse.coo_array
        disp["csc"] = scipy.sparse.csc_array
        disp["csr"] = scipy.sparse.csr_array
        fmt_l = fmt.lower()
        if fmt_l in disp:
            adj_list = self.to_adjlist(drop_islands=False)
            data = adj_list.weight
            row = adj_list.focal
            col = adj_list.neighbor
            le = _LabelEncoder()
            le.fit(row)
            row = le.transform(row)
            col = le.transform(col)
            n = self.n
            return disp[fmt_l]((data, (row, col)), shape=(n, n))
        else:
            raise ValueError(f"unsupported sparse format: {fmt}")

    @property
    def n_components(self):
        """Store whether the adjacency matrix is fully connected."""
        if "n_components" not in self._cache:
            self._n_components, self._component_labels = connected_components(
                self.sparse
            )
            self._cache["n_components"] = self._n_components
            self._cache["component_labels"] = self._component_labels
        return self._n_components

    @property
    def component_labels(self):
        """Store the graph component in which each observation falls."""
        if "component_labels" not in self._cache:
            self._n_components, self._component_labels = connected_components(
                self.sparse
            )
            self._cache["n_components"] = self._n_components
            self._cache["component_labels"] = self._component_labels
        return self._component_labels

    def _build_sparse(self):
        """Construct the sparse attribute."""

        row = []
        col = []
        data = []
        id2i = self.id2i
        for i, neigh_list in list(self.neighbor_offsets.items()):
            card = self.cardinalities[i]
            row.extend([id2i[i]] * card)
            col.extend(neigh_list)
            data.extend(self.weights[i])
        row = np.array(row)
        col = np.array(col)
        data = np.array(data)
        s = scipy.sparse.csr_matrix((data, (row, col)), shape=(self.n, self.n))
        return s

    @property
    def id2i(self):
        """Dictionary where the key is an ID and the value is that ID's
        index in ``W.id_order``.
        """
        if "id2i" not in self._cache:
            self._id2i = {}
            for i, id_i in enumerate(self._id_order):
                self._id2i[id_i] = i
            self._id2i = self._id2i
            self._cache["id2i"] = self._id2i
        return self._id2i

    @property
    def n(self):
        """Number of units."""
        if "n" not in self._cache:
            self._n = len(self.neighbors)
            self._cache["n"] = self._n
        return self._n

    @property
    def s0(self):
        r"""``s0`` is defined as

        .. math::

               s0=\sum_i \sum_j w_{i,j}

        """
        if "s0" not in self._cache:
            self._s0 = self.sparse.sum()
            self._cache["s0"] = self._s0
        return self._s0

    @property
    def s1(self):
        r"""``s1`` is defined as

        .. math::

               s1=1/2 \sum_i \sum_j \Big(w_{i,j} + w_{j,i}\Big)^2

        """
        if "s1" not in self._cache:
            t = self.sparse.transpose()
            t = t + self.sparse
            t2 = t.multiply(t)  # element-wise square
            self._s1 = t2.sum() / 2.0
            self._cache["s1"] = self._s1
        return self._s1

    @property
    def s2array(self):
        """Individual elements comprising ``s2``.

        See Also
        --------
        s2
        """
        if "s2array" not in self._cache:
            s = self.sparse
            self._s2array = np.array(s.sum(1) + s.sum(0).transpose()) ** 2
            self._cache["s2array"] = self._s2array
        return self._s2array

    @property
    def s2(self):
        r"""``s2`` is defined as

        .. math::

                s2=\sum_j \Big(\sum_i w_{i,j} + \sum_i w_{j,i}\Big)^2

        """
        if "s2" not in self._cache:
            self._s2 = self.s2array.sum()
            self._cache["s2"] = self._s2
        return self._s2

    @property
    def trcW2(self):
        """Trace of :math:`WW`.

        See Also
        --------
        diagW2
        """
        if "trcW2" not in self._cache:
            self._trcW2 = self.diagW2.sum()
            self._cache["trcw2"] = self._trcW2
        return self._trcW2

    @property
    def diagW2(self):
        """Diagonal of :math:`WW`.

        See Also
        --------
        trcW2
        """
        if "diagw2" not in self._cache:
            self._diagW2 = (self.sparse * self.sparse).diagonal()
            self._cache["diagW2"] = self._diagW2
        return self._diagW2

    @property
    def diagWtW(self):
        """Diagonal of :math:`W^{'}W`.

        See Also
        --------
        trcWtW
        """
        if "diagWtW" not in self._cache:
            self._diagWtW = (self.sparse.transpose() * self.sparse).diagonal()
            self._cache["diagWtW"] = self._diagWtW
        return self._diagWtW

    @property
    def trcWtW(self):
        """Trace of :math:`W^{'}W`.

        See Also
        --------
        diagWtW
        """
        if "trcWtW" not in self._cache:
            self._trcWtW = self.diagWtW.sum()
            self._cache["trcWtW"] = self._trcWtW
        return self._trcWtW

    @property
    def diagWtW_WW(self):
        """Diagonal of :math:`W^{'}W + WW`."""
        if "diagWtW_WW" not in self._cache:
            wt = self.sparse.transpose()
            w = self.sparse
            self._diagWtW_WW = (wt * w + w * w).diagonal()
            self._cache["diagWtW_WW"] = self._diagWtW_WW
        return self._diagWtW_WW

    @property
    def trcWtW_WW(self):
        """Trace of :math:`W^{'}W + WW`."""
        if "trcWtW_WW" not in self._cache:
            self._trcWtW_WW = self.diagWtW_WW.sum()
            self._cache["trcWtW_WW"] = self._trcWtW_WW
        return self._trcWtW_WW

    @property
    def pct_nonzero(self):
        """Percentage of nonzero weights."""
        if "pct_nonzero" not in self._cache:
            self._pct_nonzero = 100.0 * self.sparse.nnz / (1.0 * self._n**2)
            self._cache["pct_nonzero"] = self._pct_nonzero
        return self._pct_nonzero

    @property
    def cardinalities(self):
        """Number of neighbors for each observation."""
        if "cardinalities" not in self._cache:
            c = {}
            for i in self._id_order:
                c[i] = len(self.neighbors[i])
            self._cardinalities = c
            self._cache["cardinalities"] = self._cardinalities
        return self._cardinalities

    @property
    def max_neighbors(self):
        """Largest number of neighbors."""
        if "max_neighbors" not in self._cache:
            self._max_neighbors = max(self.cardinalities.values())
            self._cache["max_neighbors"] = self._max_neighbors
        return self._max_neighbors

    @property
    def mean_neighbors(self):
        """Average number of neighbors."""
        if "mean_neighbors" not in self._cache:
            self._mean_neighbors = np.mean(list(self.cardinalities.values()))
            self._cache["mean_neighbors"] = self._mean_neighbors
        return self._mean_neighbors

    @property
    def min_neighbors(self):
        """Minimum number of neighbors."""
        if "min_neighbors" not in self._cache:
            self._min_neighbors = min(self.cardinalities.values())
            self._cache["min_neighbors"] = self._min_neighbors
        return self._min_neighbors

    @property
    def nonzero(self):
        """Number of nonzero weights."""
        if "nonzero" not in self._cache:
            self._nonzero = self.sparse.nnz
            self._cache["nonzero"] = self._nonzero
        return self._nonzero

    @property
    def sd(self):
        """Standard deviation of number of neighbors."""
        if "sd" not in self._cache:
            self._sd = np.std(list(self.cardinalities.values()))
            self._cache["sd"] = self._sd
        return self._sd

    @property
    def asymmetries(self):
        """List of id pairs with asymmetric weights
        sorted in ascending *index location* order.
        """
        if "asymmetries" not in self._cache:
            self._asymmetries = self.asymmetry()
            self._cache["asymmetries"] = self._asymmetries
        return self._asymmetries

    @property
    def islands(self):
        """List of ids without any neighbors."""
        if "islands" not in self._cache:
            self._islands = [i for i, c in list(self.cardinalities.items()) if c == 0]
            self._cache["islands"] = self._islands
        return self._islands

    @property
    def histogram(self):
        """Cardinality histogram as a dictionary where key is the id and
        value is the number of neighbors for that unit.
        """
        if "histogram" not in self._cache:
            ct, bin_ = np.histogram(
                list(self.cardinalities.values()),
                list(range(self.min_neighbors, self.max_neighbors + 2)),
            )
            self._histogram = list(zip(bin_[:-1], ct, strict=True))
            self._cache["histogram"] = self._histogram
        return self._histogram

    def __getitem__(self, key):
        """Allow a dictionary like interaction with the weights class.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w = lat2W()

        >>> w[0] == dict({1: 1.0, 5: 1.0})
        True
        """
        return dict(list(zip(self.neighbors[key], self.weights[key], strict=True)))

    def __iter__(self):
        """
        Support iteration over weights.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W(3,3)
        >>> for i,wi in enumerate(w):
        ...     print(i,wi[0])
        ...
        0 0
        1 1
        2 2
        3 3
        4 4
        5 5
        6 6
        7 7
        8 8
        >>>
        """
        for i in self._id_order:
            yield i, dict(list(zip(self.neighbors[i], self.weights[i], strict=True)))

    def remap_ids(self, new_ids):
        """
        In place modification throughout ``W`` of id values from
        ``w.id_order`` to ``new_ids`` in all.

        Parameters
        ----------
        new_ids : list, numpy.ndarray
            Aligned list of new ids to be inserted. Note that first
            element of ``new_ids`` will replace first element of
            ``w.id_order``, second element of ``new_ids`` replaces second
            element of ``w.id_order`` and so on.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w = lat2W(3, 3)
        >>> w.id_order
        [0, 1, 2, 3, 4, 5, 6, 7, 8]
        >>> w.neighbors[0]
        [3, 1]
        >>> new_ids = ['id%i'%id for id in w.id_order]
        >>> _ = w.remap_ids(new_ids)
        >>> w.id_order
        ['id0', 'id1', 'id2', 'id3', 'id4', 'id5', 'id6', 'id7', 'id8']
        >>> w.neighbors['id0']
        ['id3', 'id1']
        """

        old_ids = self._id_order
        if len(old_ids) != len(new_ids):
            raise Exception(
                "W.remap_ids: length of `old_ids` does not match             that of"
                " new_ids"
            )
        if len(set(new_ids)) != len(new_ids):
            raise Exception("W.remap_ids: list `new_ids` contains duplicates")
        else:
            new_neighbors = {}
            new_weights = {}
            old_transformations = self.transformations["O"].copy()
            new_transformations = {}
            for o, n in zip(old_ids, new_ids, strict=True):
                o_neighbors = self.neighbors[o]
                o_weights = self.weights[o]
                n_neighbors = [new_ids[old_ids.index(j)] for j in o_neighbors]
                new_neighbors[n] = n_neighbors
                new_weights[n] = o_weights[:]
                new_transformations[n] = old_transformations[o]
            self.neighbors = new_neighbors
            self.weights = new_weights
            self.transformations["O"] = new_transformations

            id_order = [self._id_order.index(o) for o in old_ids]
            for i, id_ in enumerate(id_order):
                self.id_order[id_] = new_ids[i]

            self._reset()

    def __set_id_order(self, ordered_ids):
        """Set the iteration order in w. ``W`` can be iterated over. On
        construction the iteration order is set to the lexicographic order of
        the keys in the ``w.weights`` dictionary. If a specific order
        is required it can be set with this method.

        Parameters
        ----------
        ordered_ids : sequence
            Identifiers for observations in specified order.

        Notes
        -----
        The ``ordered_ids`` parameter is checked against the ids implied
        by the keys in ``w.weights``. If they are not equivalent sets an
        exception is raised and the iteration order is not changed.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W(3,3)
        >>> for i,wi in enumerate(w):
        ...     print(i, wi[0])
        ...
        0 0
        1 1
        2 2
        3 3
        4 4
        5 5
        6 6
        7 7
        8 8
        >>> w.id_order
        [0, 1, 2, 3, 4, 5, 6, 7, 8]
        >>> w.id_order=range(8,-1,-1)
        >>> list(w.id_order)
        [8, 7, 6, 5, 4, 3, 2, 1, 0]
        >>> for i,w_i in enumerate(w):
        ...     print(i,w_i[0])
        ...
        0 8
        1 7
        2 6
        3 5
        4 4
        5 3
        6 2
        7 1
        8 0
        """

        if set(self._id_order) == set(ordered_ids):
            self._id_order = ordered_ids
            self._id_order_set = True
            self._reset()
        else:
            raise Exception("ordered_ids do not align with W ids")

    def __get_id_order(self):
        """Returns the ids for the observations in the order in which they
        would be encountered if iterating over the weights.
        """
        return self._id_order

    id_order = property(__get_id_order, __set_id_order)

    @property
    def id_order_set(self):
        """Returns ``True`` if user has set ``id_order``, ``False`` if not.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W()
        >>> w.id_order_set
        True
        """
        return self._id_order_set

    @property
    def neighbor_offsets(self):
        """
        Given the current ``id_order``, ``neighbor_offsets[id]`` is the
        offsets of the id's neighbors in ``id_order``.

        Returns
        -------
        neighbor_list : list
            Offsets of the id's neighbors in ``id_order``.

        Examples
        --------
        >>> from libpysal.weights import W
        >>> neighbors={'c': ['b'], 'b': ['c', 'a'], 'a': ['b']}
        >>> weights ={'c': [1.0], 'b': [1.0, 1.0], 'a': [1.0]}
        >>> w=W(neighbors,weights)
        >>> w.id_order = ['a','b','c']
        >>> w.neighbor_offsets['b']
        [2, 0]
        >>> w.id_order = ['b','a','c']
        >>> w.neighbor_offsets['b']
        [2, 1]
        """

        if "neighbors_0" not in self._cache:
            self.__neighbors_0 = {}
            id2i = self.id2i
            for j, neigh_list in list(self.neighbors.items()):
                self.__neighbors_0[j] = [id2i[neigh] for neigh in neigh_list]
            self._cache["neighbors_0"] = self.__neighbors_0

        neighbor_list = self.__neighbors_0

        return neighbor_list

    def get_transform(self):
        """Getter for transform property.

        Returns
        -------
        transformation : str, None
            Valid transformation value. See the ``transform``
            parameters in ``set_transform()`` for a detailed description.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W()
        >>> w.weights[0]
        [1.0, 1.0]
        >>> w.transform
        'O'
        >>> w.transform='r'
        >>> w.weights[0]
        [0.5, 0.5]
        >>> w.transform='b'
        >>> w.weights[0]
        [1.0, 1.0]

        See also
        --------
        set_transform
        """

        return self._transform

    def set_transform(self, value="B"):
        """Transformations of weights.

        Parameters
        ----------
        transform : str
            This parameter is not case sensitive. The following are
            valid transformations.

            * **B** -- Binary
            * **R** -- Row-standardization (global sum :math:`=n`)
            * **D** -- Double-standardization (global sum :math:`=1`)
            * **V** -- Variance stabilizing
            * **O** -- Restore original transformation (from instantiation)

        Notes
        -----
        Transformations are applied only to the value of the weights at
        instantiation. Chaining of transformations cannot be done on a ``W``
        instance.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W()
        >>> w.weights[0]
        [1.0, 1.0]
        >>> w.transform
        'O'
        >>> w.transform='r'
        >>> w.weights[0]
        [0.5, 0.5]
        >>> w.transform='b'
        >>> w.weights[0]
        [1.0, 1.0]
        """
        value = value.upper()
        self._transform = value
        if value in self.transformations:
            self.weights = self.transformations[value]
            self._reset()
        else:
            if value == "R":
                # row standardized weights
                weights = {}
                self.weights = self.transformations["O"]
                for i in self.weights:
                    wijs = self.weights[i]
                    row_sum = sum(wijs) * 1.0
                    if row_sum == 0.0 and not self.silence_warnings:
                        print(("WARNING: ", i, " is an island (no neighbors)"))
                    weights[i] = [wij / row_sum for wij in wijs]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "D":
                # doubly-standardized weights
                # update current chars before doing global sum
                self._reset()
                s0 = self.s0
                ws = 1.0 / s0
                weights = {}
                self.weights = self.transformations["O"]
                for i in self.weights:
                    wijs = self.weights[i]
                    weights[i] = [wij * ws for wij in wijs]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "B":
                # binary transformation
                weights = {}
                self.weights = self.transformations["O"]
                for i in self.weights:
                    wijs = self.weights[i]
                    weights[i] = [1.0 for wij in wijs]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "V":
                # variance stabilizing
                weights = {}
                q = {}
                s = {}
                big_q = 0.0
                self.weights = self.transformations["O"]
                for i in self.weights:
                    wijs = self.weights[i]
                    q[i] = math.sqrt(sum([wij * wij for wij in wijs]))
                    s[i] = [wij / q[i] for wij in wijs]
                    big_q += sum(s[i])
                nq = self.n / big_q
                for i in self.weights:
                    weights[i] = [w * nq for w in s[i]]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "O":
                # put weights back to original transformation
                weights = {}
                original = self.transformations[value]
                self.weights = original
                self._reset()
            else:
                raise Exception("unsupported weights transformation")

    transform = property(get_transform, set_transform)

    def asymmetry(self, intrinsic=True):
        r"""
        Asymmetry check.

        Parameters
        ----------
        intrinsic : bool
            Default is ``True``. Intrinsic symmetry is defined as:

            .. math::

                w_{i,j} == w_{j,i}

            If ``intrinsic`` is ``False`` symmetry is defined as:

            .. math::

                i \in N_j \ \& \ j \in N_i

            where :math:`N_j` is the set of neighbors for :math:`j`.

        Returns
        -------
        asymmetries : list
            Empty if no asymmetries are found. If there are asymmetries,
            then a ``list`` of ``(i,j)`` tuples is returned sorted in
            ascending *index location* order.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W(3,3)
        >>> w.asymmetry()
        []
        >>> w.transform='r'
        >>> w.asymmetry()
        [(0, 1), (0, 3), (1, 0), (1, 2), (1, 4), (2, 1), (2, 5), (3, 0), (3, 4), (3, 6), (4, 1), (4, 3), (4, 5), (4, 7), (5, 2), (5, 4), (5, 8), (6, 3), (6, 7), (7, 4), (7, 6), (7, 8), (8, 5), (8, 7)]
        >>> result = w.asymmetry(intrinsic=False)
        >>> result
        []
        >>> neighbors={0:[1,2,3], 1:[1,2,3], 2:[0,1], 3:[0,1]}
        >>> weights={0:[1,1,1], 1:[1,1,1], 2:[1,1], 3:[1,1]}
        >>> w=W(neighbors,weights)
        >>> w.asymmetry()
        [(0, 1), (1, 0)]
        """  # noqa: E501

        if intrinsic:
            wd = self.sparse.transpose() - self.sparse
        else:
            transform = self.transform
            self.transform = "b"
            wd = self.sparse.transpose() - self.sparse
            self.transform = transform

        ids = np.nonzero(wd)

        if len(ids[0]) == 0:
            return []
        else:
            ijs = list(zip(ids[0], ids[1], strict=True))
            ijs.sort()

            i2id = {v: k for k, v in self.id2i.items()}
            ijs = [(i2id[i], i2id[j]) for i, j in ijs]

            return ijs

    def symmetrize(self, inplace=False):
        """Construct a symmetric KNN weight. This ensures that the neighbors
        of each focal observation consider the focal observation itself as
        a neighbor. This returns a generic ``W`` object, since the object is no
        longer guaranteed to have ``k`` neighbors for each observation.
        """
        if not inplace:
            neighbors = copy.deepcopy(self.neighbors)
            weights = copy.deepcopy(self.weights)
            out_w = W(neighbors, weights, id_order=self.id_order)
            out_w.symmetrize(inplace=True)
            return out_w
        else:
            for focal, fneighbs in list(self.neighbors.items()):
                for j, neighbor in enumerate(fneighbs):
                    neighb_neighbors = self.neighbors[neighbor]
                    if focal not in neighb_neighbors:
                        self.neighbors[neighbor].append(focal)
                        self.weights[neighbor].append(self.weights[focal][j])
            self._cache = {}
            return

    def full(self):
        """Generate a full ``numpy.ndarray``.

        Parameters
        ----------
        self : libpysal.weights.W
            spatial weights object

        Returns
        -------
        (fullw, keys) : tuple
            The first element being the full ``numpy.ndarray`` and second
            element keys being the ids associated with each row in the array.

        Examples
        --------
        >>> from libpysal.weights import W, full
        >>> neighbors = {
        ...     'first':['second'],'second':['first','third'],'third':['second']
        ... }
        >>> weights = {'first':[1],'second':[1,1],'third':[1]}
        >>> w = W(neighbors, weights)
        >>> wf, ids = full(w)
        >>> wf
        array([[0., 1., 0.],
               [1., 0., 1.],
               [0., 1., 0.]])
        >>> ids
        ['first', 'second', 'third']
        """
        wfull = self.sparse.toarray()
        keys = list(self.neighbors.keys())
        if self.id_order:
            keys = self.id_order

        return (wfull, keys)

    def to_WSP(self):
        """Generate a ``WSP`` object.

        Returns
        -------
        implicit : libpysal.weights.WSP
            Thin ``W`` class

        Examples
        --------
        >>> from libpysal.weights import W, WSP
        >>> neighbors={'first':['second'],'second':['first','third'],'third':['second']}
        >>> weights={'first':[1],'second':[1,1],'third':[1]}
        >>> w=W(neighbors,weights)
        >>> wsp=w.to_WSP()
        >>> isinstance(wsp, WSP)
        True
        >>> wsp.n
        3
        >>> wsp.s0
        4

        See also
        --------
        WSP
        """
        return WSP(self.sparse, self._id_order)

    def set_shapefile(self, shapefile, idVariable=None, full=False):
        """
        Adding metadata for writing headers of ``.gal`` and ``.gwt`` files.

        Parameters
        ----------
        shapefile : str
            The shapefile name used to construct weights.
        idVariable : str
            The name of the attribute in the shapefile to associate
            with ids in the weights.
        full : bool
            Write out the entire path for a shapefile (``True``) or
            only the base of the shapefile without extension (``False``).
            Default is ``True``.
        """

        if full:
            self._shpName = shapefile
        else:
            self._shpName = basename(shapefile).split(".")[0]

        self._varName = idVariable

    def plot(
        self, gdf, indexed_on=None, ax=None, color="k", node_kws=None, edge_kws=None
    ):
        """Plot spatial weights objects. **Requires** ``matplotlib``, and
        implicitly requires a ``geopandas.GeoDataFrame`` as input.

        Parameters
        ----------
        gdf : geopandas.GeoDataFrame
            The original shapes whose topological relations are modelled in ``W``.
        indexed_on : str
            Column of ``geopandas.GeoDataFrame`` that the weights object uses
            as an index. Default is ``None``, so the index of the
            ``geopandas.GeoDataFrame`` is used.
        ax : matplotlib.axes.Axes
            Axis on which to plot the weights. Default is ``None``, so
            plots on the current figure.
        color : str
            ``matplotlib`` color string, will color both nodes and edges
            the same by default.
        node_kws : dict
            Keyword arguments dictionary to send to ``pyplot.scatter``,
            which provides fine-grained control over the aesthetics
            of the nodes in the plot.
        edge_kws : dict
            Keyword arguments dictionary to send to ``pyplot.plot``,
            which provides fine-grained control over the aesthetics
            of the edges in the plot.

        Returns
        -------
        f : matplotlib.figure.Figure
            Figure on which the plot is made.
        ax : matplotlib.axes.Axes
            Axis on which the plot is made.

        Notes
        -----
        If you'd like to overlay the actual shapes from the
        ``geopandas.GeoDataFrame``, call ``gdf.plot(ax=ax)`` after this.
        To plot underneath, adjust the z-order of the plot as follows:
        ``gdf.plot(ax=ax,zorder=0)``.

        Examples
        --------
        >>> from libpysal.weights import Queen
        >>> import libpysal as lp
        >>> import geopandas
        >>> gdf = geopandas.read_file(lp.examples.get_path("columbus.shp"))
        >>> weights = Queen.from_dataframe(gdf)
        >>> tmp = weights.plot(
        ...     gdf, color='firebrickred', node_kws=dict(marker='*', color='k')
        ... )
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError(
                "W.plot depends on matplotlib.pyplot, and this was"
                "not able to be imported. \nInstall matplotlib to"
                "plot spatial weights."
            ) from None
        if ax is None:
            f = plt.figure()
            ax = plt.gca()
        else:
            f = plt.gcf()
        if node_kws is not None:
            if "color" not in node_kws:
                node_kws["color"] = color
        else:
            node_kws = {"color": color}
        if edge_kws is not None:
            if "color" not in edge_kws:
                edge_kws["color"] = color
        else:
            edge_kws = {"color": color}

        for idx, neighbors in self.neighbors.items():
            if idx in self.islands:
                continue
            if indexed_on is not None:
                neighbors = gdf[gdf[indexed_on].isin(neighbors)].index.tolist()
                idx = gdf[gdf[indexed_on] == idx].index.tolist()[0]
            else:
                neighbors = list(neighbors)
            centroids = gdf.loc[neighbors].centroid
            centroids = np.stack([centroids.x, centroids.y], axis=1)
            focal = np.hstack(gdf.loc[idx].geometry.centroid.xy)
            seen = set()
            for nidx, neighbor in zip(neighbors, centroids, strict=True):
                if (idx, nidx) in seen:
                    continue
                ax.plot(
                    *list(zip(focal, neighbor, strict=True)), marker=None, **edge_kws
                )
                seen.update((idx, nidx))
                seen.update((nidx, idx))
        centroids = gdf.centroid
        ax.scatter(centroids.x, centroids.y, **node_kws)
        return f, ax


class WSP:
    """Thin ``W`` class for ``spreg``.

    Parameters
    ----------
    sparse : scipy.sparse.{matrix-type}
        NxN object from ``scipy.sparse``

    Attributes
    ----------
    n           : int
                  description
    s0          : float
                  description
    trcWtW_WW   : float
                  description

    Examples
    --------
    From GAL information

    >>> import scipy.sparse
    >>> from libpysal.weights import WSP
    >>> rows = [0, 1, 1, 2, 2, 3]
    >>> cols = [1, 0, 2, 1, 3, 3]
    >>> weights =  [1, 0.75, 0.25, 0.9, 0.1, 1]
    >>> sparse = scipy.sparse.csr_matrix((weights, (rows, cols)), shape=(4,4))
    >>> w = WSP(sparse)
    >>> w.s0
    4.0
    >>> w.trcWtW_WW
    6.395
    >>> w.n
    4
    """

    def __init__(self, sparse, id_order=None, index=None):
        if not scipy.sparse.issparse(sparse):
            raise ValueError("must pass a scipy sparse object")
        rows, cols = sparse.shape
        if rows != cols:
            raise ValueError("Weights object must be square")
        self.sparse = sparse.tocsr()
        self.n = sparse.shape[0]
        self._cache = {}
        if id_order:
            if len(id_order) != self.n:
                raise ValueError(
                    "Number of values in id_order must match shape of sparse"
                )
            else:
                self._id_order = id_order
                self._cache["id_order"] = self._id_order
        # temp addition of index attribute
        import pandas as pd  # will be removed after refactoring is done

        if index is not None:
            if not isinstance(index, pd.Index | pd.MultiIndex | pd.RangeIndex):
                raise TypeError("index must be an instance of pandas.Index dtype")
            if len(index) != self.n:
                raise ValueError("Number of values in index must match shape of sparse")
        else:
            index = pd.RangeIndex(self.n)
        self.index = index

    @property
    def id_order(self):
        """An ordered list of ids, assumed to match the ordering in ``sparse``."""
        # Temporary solution until the refactoring is finished
        if "id_order" not in self._cache:
            if hasattr(self, "index"):
                self._id_order = self.index.tolist()
            else:
                self._id_order = list(range(self.n))
            self._cache["id_order"] = self._id_order
        return self._id_order

    @property
    def s0(self):
        r"""``s0`` is defined as:

        .. math::

               s0=\sum_i \sum_j w_{i,j}

        """
        if "s0" not in self._cache:
            self._s0 = self.sparse.sum()
            self._cache["s0"] = self._s0
        return self._s0

    @property
    def trcWtW_WW(self):
        """Trace of :math:`W^{'}W + WW`."""
        if "trcWtW_WW" not in self._cache:
            self._trcWtW_WW = self.diagWtW_WW.sum()
            self._cache["trcWtW_WW"] = self._trcWtW_WW
        return self._trcWtW_WW

    @property
    def diagWtW_WW(self):
        """Diagonal of :math:`W^{'}W + WW`."""
        if "diagWtW_WW" not in self._cache:
            wt = self.sparse.transpose()
            w = self.sparse
            self._diagWtW_WW = (wt * w + w * w).diagonal()
            self._cache["diagWtW_WW"] = self._diagWtW_WW
        return self._diagWtW_WW

    @classmethod
    def from_W(cls, W):
        """Constructs a ``WSP`` object from the ``W``'s sparse matrix.

        Parameters
        ----------
        W : libpysal.weights.W
            A PySAL weights object with a sparse form and ids.

        Returns
        -------
        A ``WSP`` instance.
        """
        return cls(W.sparse, id_order=W.id_order)

    def to_W(self, silence_warnings=False):
        """
        Convert a pysal WSP object (thin weights matrix) to a pysal W object.

        Parameters
        ----------
        self : WSP
            PySAL sparse weights object.
        silence_warnings : bool
            Switch to ``True`` to turn off print statements for every
            observation with islands. Default is ``False``, which does
            not silence warnings.

        Returns
        -------
        w : W
            PySAL weights object.

        Examples
        --------
        >>> from libpysal.weights import lat2SW, WSP, WSP2W

        Build a 10x10 ``scipy.sparse`` matrix for a rectangular 2x5
        region of cells (rook contiguity), then construct a ``libpysal``
        sparse weights object (``self``).

        >>> sp = lat2SW(2, 5)
        >>> self = WSP(sp)
        >>> self.n
        10
        >>> print(self.sparse[0].todense())
        [[0 1 0 0 0 1 0 0 0 0]]

        Convert this sparse weights object to a standard PySAL weights object.

        >>> w = WSP2W(self)
        >>> w.n
        10
        >>> print(w.full()[0][0])
        [0. 1. 0. 0. 0. 1. 0. 0. 0. 0.]
        """

        indices = list(self.sparse.indices)
        data = list(self.sparse.data)
        indptr = list(self.sparse.indptr)
        id_order = self.id_order
        if id_order:
            # replace indices with user IDs
            indices = [id_order[i] for i in indices]
        else:
            id_order = list(range(self.n))
        neighbors, weights = {}, {}
        start = indptr[0]
        for i in range(self.n):
            oid = id_order[i]
            end = indptr[i + 1]
            neighbors[oid] = indices[start:end]
            weights[oid] = data[start:end]
            start = end
        ids = copy.copy(self.id_order)
        w = W(neighbors, weights, ids, silence_warnings=silence_warnings)
        w._sparse = copy.deepcopy(self.sparse)
        w._cache["sparse"] = w._sparse
        return w



import os
import tempfile

import pytest

from .... import examples as pysal_examples
from ...fileio import FileIO
from ..mtx import MtxIO


class TesttestMtxIO:
    def setup_method(self):
        self.test_file = test_file = pysal_examples.get_path("wmat.mtx")
        self.obj = MtxIO(test_file, "r")

    def test_close(self):
        f = self.obj
        f.close()
        pytest.raises(ValueError, f.read)

    def test_read(self):
        w = self.obj.read()
        assert w.n == 49
        assert w.mean_neighbors == 4.7346938775510203
        assert list(w[1].values()) == [
            0.33329999999999999,
            0.33329999999999999,
            0.33329999999999999,
        ]
        s0 = w.s0
        self.obj.seek(0)
        wsp = self.obj.read(sparse=True)
        assert wsp.n == 49
        assert s0 == wsp.s0

    def test_seek(self):
        self.test_read()
        pytest.raises(StopIteration, self.obj.read)
        self.obj.seek(0)
        self.test_read()

    def test_write(self):
        for i in [False, True]:
            self.obj.seek(0)
            w = self.obj.read(sparse=i)
            f = tempfile.NamedTemporaryFile(suffix=".mtx")
            fname = f.name
            f.close()
            o = FileIO(fname, "w")
            o.write(w)
            o.close()
            wnew = FileIO(fname, "r").read(sparse=i)
            if i:
                assert wnew.s0 == w.s0
            else:
                assert wnew.pct_nonzero == w.pct_nonzero
            os.remove(fname)



from collections import defaultdict

import geopandas
import numpy
import pandas
import shapely
from packaging.version import Version

from ._utils import _neighbor_dict_to_edges, _resolve_islands, _validate_geometry_input

GPD_013 = Version(geopandas.__version__) >= Version("0.13")

_VALID_GEOMETRY_TYPES = ["Polygon", "MultiPolygon", "LineString", "MultiLineString"]


def _vertex_set_intersection(geoms, rook=True, ids=None, by_perimeter=False):
    """
    Use a hash map inversion to construct a graph

    Parameters
    ---------
    geoms : geopandas.GeoDataFrame, geopandas.GeoSeries, numpy.array
        The container for the geometries to compute contiguity. Regardless of
        the containing type, the geometries within the container must be Polygons
        or MultiPolygons.
    rook : bool (default: True)
        whether to compute vertex set intersection contiguity by edge or by point.
        By default, vertex set contiguity is computed by edge. This means that at least
        two adjacent vertices on the polygon boundary must be shared.
    ids : numpy.ndarray (default: None)
        names to use for indexing the graph constructed from geoms. If None (default),
        an index is extracted from `geoms`. If `geoms` has no index, a pandas.RangeIndex
        is constructed.
    by_perimeter : bool (default: False)
        whether to compute perimeter-weighted contiguity. By default, this returns
        the raw length of perimeter overlap betwen contiguous polygons or lines.
        In the case of LineString/MultiLineString input geoms, this is likely to
        result in empty weights, where all observations are isolates.

    """
    _, ids, geoms = _validate_geometry_input(
        geoms, ids=ids, valid_geometry_types=_VALID_GEOMETRY_TYPES
    )

    # initialise the target map
    graph = {}
    for idx in ids:
        graph[idx] = {idx}

    # get all of the vertices for the input
    vertices, offsets = shapely.get_coordinates(geoms.geometry, return_index=True)
    # use offsets from exploded geoms to create edges to avoid a phantom edge between
    # parts of multipolygon
    _, single_part_offsets = shapely.get_coordinates(
        geoms.geometry.explode(ignore_index=True), return_index=True
    )
    # initialise the hashmap we want to invert
    vert_to_geom = defaultdict(set)

    # populate the hashmap we intend to invert
    if rook:
        for i, vertex in enumerate(vertices[:-1]):
            if single_part_offsets[i] != single_part_offsets[i + 1]:
                continue
            edge = tuple(sorted([tuple(vertex), tuple(vertices[i + 1])]))
            # edge to {polygons, with, this, edge}
            vert_to_geom[edge].add(offsets[i])
    else:
        for i, vertex in enumerate(vertices):
            # vertex to {polygons, with, this, vertex}
            vert_to_geom[tuple(vertex)].add(offsets[i])

    # invert vert_to_geom
    for nexus in vert_to_geom.values():
        if len(nexus) < 2:
            continue
        nexus_names = {ids[ix] for ix in nexus}
        for geom_ix in nexus:
            gid = ids[geom_ix]
            graph[gid].update(nexus_names)

    for idx in ids:
        graph[idx].remove(idx)

    heads, tails, weights = _neighbor_dict_to_edges(graph)

    if by_perimeter:
        weights = numpy.zeros(len(heads), dtype=float)
        non_isolates = heads != tails  # can't pass isolates to _perimeter_weigths
        weights[non_isolates] = _perimeter_weights(
            geoms, heads[non_isolates], tails[non_isolates]
        )

    return heads, tails, weights


def _queen(geoms, ids=None, by_perimeter=False):
    """
    Construct queen contiguity using point-set relations.

    Queen contiguity occurs when two polygons touch at exactly a point.
    Overlapping polygons will not be considered as neighboring
    under this rule, since contiguity is strictly planar.

    Parameters
    ----------
    geoms : geopandas.GeoDataFrame, geopandas.GeoSeries, numpy.array
        The container for the geometries to compute contiguity. Regardless of
        the containing type, the geometries within the container must be Polygons
        or MultiPolygons.
    ids : numpy.ndarray (default: None)
        names to use for indexing the graph constructed from geoms. If None (default),
        an index is extracted from `geoms`. If `geoms` has no index, a pandas.RangeIndex
        is constructed.
    by_perimeter : bool (default: False)
        whether to compute perimeter-weighted contiguity. By default, this returns
        the raw length of perimeter overlap betwen contiguous polygons or lines.
        In the case of LineString/MultiLineString input geoms, this is likely to
        result in empty weights, where all observations are isolates.

    Returns
    -------
    (heads, tails, weights) : three vectors describing the links in the
        queen contiguity graph, with islands represented as a self-loop with
        zero weight.
    """
    _, ids, geoms = _validate_geometry_input(
        geoms, ids=ids, valid_geometry_types=_VALID_GEOMETRY_TYPES
    )
    heads_ix, tails_ix = shapely.STRtree(geoms).query(geoms, predicate="touches")
    heads, tails = ids[heads_ix], ids[tails_ix]

    if by_perimeter:
        weights = _perimeter_weights(geoms, heads, tails)
    else:
        weights = numpy.ones_like(heads_ix, dtype=int)

    return _resolve_islands(heads, tails, ids, weights=weights)


def _rook(geoms, ids=None, by_perimeter=False):
    """
    Construct rook contiguity using point-set relations.

    Rook contiguity occurs when two polygons touch over at least one edge.
    Overlapping polygons will not be considered as neighboring
    under this rule, since contiguity is strictly planar.

    Parameters
    ----------
    geoms : geopandas.GeoDataFrame, geopandas.GeoSeries, numpy.array
        The container for the geometries to compute contiguity. Regardless of
        the containing type, the geometries within the container must be Polygons
        or MultiPolygons.
    ids : numpy.ndarray (default: None)
        names to use for indexing the graph constructed from geoms. If None (default),
        an index is extracted from `geoms`. If `geoms` has no index, a pandas.RangeIndex
        is constructed.
    by_perimeter : bool (default: False)
        whether to compute perimeter-weighted contiguity. By default, this returns
        the raw length of perimeter overlap betwen contiguous polygons or lines.
        In the case of LineString/MultiLineString input geoms, this is likely to
        result in empty weights, where all observations are isolates.

    Returns
    -------
    (heads, tails, weights) : three vectors describing the links in the
        rook contiguity graph, with islands represented as a self-loop with
        zero weight.
    """
    _, ids, geoms = _validate_geometry_input(
        geoms, ids=ids, valid_geometry_types=_VALID_GEOMETRY_TYPES
    )
    heads_ix, tails_ix = shapely.STRtree(geoms).query(geoms)
    mask = shapely.relate_pattern(
        geoms.values[heads_ix], geoms.values[tails_ix], "F***1****"
    )
    heads, tails = ids[heads_ix][mask], ids[tails_ix][mask]
    if by_perimeter:
        weights = _perimeter_weights(geoms, heads, tails)

    if not by_perimeter:
        weights = numpy.ones_like(heads, dtype=int)

    return _resolve_islands(heads, tails, ids, weights)


def _perimeter_weights(geoms, heads, tails):
    """
    Compute the perimeter of neighbor pairs for edges describing a contiguity graph.

    Note that this result will be incorrect if the head and tail polygon overlap.
    If they do overlap, it is an "invalid" contiguity, so the length of the
    perimeter of the intersection may not express the correct value for relatedness
    in the contiguity graph.

    This is a private method, so strict conditions
    on input data are expected.
    """
    intersection = shapely.intersection(geoms[heads].values, geoms[tails].values)
    geom_types = shapely.get_type_id(shapely.get_parts(intersection))

    # check if the intersection resulted in (Multi)Polygon
    if numpy.isin(geom_types, [3, 6]).any():
        raise ValueError(
            "Some geometries overlap. Perimeter weights require planar coverage."
        )

    return shapely.length(intersection)


def _block_contiguity(regimes, ids=None):
    """Construct spatial weights for regime neighbors.

    Block contiguity structures are relevant when defining neighbor relations
    based on membership in a regime. For example, all counties belonging to
    the same state could be defined as neighbors, in an analysis of all
    counties in the US.

    Parameters
    ----------
    regimes : list-like
        list-like of regimes. If pandas.Series, its index is used to encode Graph
    ids : list-like, optional
        ordered sequence of IDs for the observations to be used as an index, by default
        None. If ``regimes`` is not a pandas.Series and ids=None, range index is used.

    Returns
    -------
    dict
        dictionary of neighbors
    """
    regimes = pandas.Series(regimes, index=ids)
    rids = regimes.unique()
    neighbors = {}
    for rid in rids:
        members = regimes.index[regimes == rid].values
        for member in members:
            neighbors[member] = members[members != member]
    return neighbors


def _fuzzy_contiguity(
    geoms, ids, tolerance=None, buffer=None, predicate="intersects", **kwargs
):
    """Fuzzy contiguity builder

    Parameters
    ----------
    geoms : array-like of shapely.Geometry objects
        Could be geopandas.GeoSeries or geopandas.GeoDataFrame, in which case ids need
        to match geoms.index.
    ids : array
        ids to be used index of the adjacency
    tolerance : float, optional
        The percentage of the length of the minimum side of the bounding rectangle
        for the ``geoms`` to use in determining the buffering distance. Either
        ``tolerance`` or ``buffer`` may be specified but not both.
        By default None.
    buffer : float, optional
        Exact buffering distance in the units of ``geoms.crs``. Either
        ``tolerance`` or ``buffer`` may be specified but not both.
        By default None.
    predicate : str, optional
        The predicate to use for determination of neighbors. Default is 'intersects'.
        If None is passed, neighbours are determined based on the intersection of
        bounding boxes. See the documentation of ``geopandas.GeoSeries.sindex.query``
        for allowed predicates.
    **kwargs
        Keyword arguments passed to ``geopandas.GeoSeries.buffer``.

    Returns
    -------
    tuple
        tuple of ``heads``, ``tails``, ``weights`` arrays
    """
    if buffer is not None and tolerance is not None:
        raise ValueError(
            "Only one of `tolerance` and `buffer` can be speciifed, not both."
        )

    if not isinstance(geoms, geopandas.base.GeoPandasBase):
        geoms = geopandas.GeoSeries(geoms, index=ids)

    if tolerance is not None:
        minx, miny, maxx, maxy = geoms.total_bounds
        buffer = tolerance * 0.5 * abs(min(maxx - minx, maxy - miny))

    if buffer is not None:
        geoms = geoms.buffer(buffer, **kwargs)

    # query tree based on set predicate
    if GPD_013:
        head, tail = geoms.sindex.query(geoms.geometry, predicate=predicate)
    else:
        head, tail = geoms.sindex.query_bulk(geoms.geometry, predicate=predicate)
    # remove self hits
    itself = head == tail
    heads = ids[head[~itself]]
    tails = ids[tail[~itself]]

    weights = numpy.ones_like(heads, dtype=int)

    return _resolve_islands(heads, tails, ids.values, weights=weights)



import geodatasets
import geopandas
import numpy as np
import pytest
import shapely

from libpysal import graph
from libpysal.graph.tests.test_utils import fetch_map_string


@pytest.mark.network
class TestPlotting:
    def setup_method(self):
        _ = pytest.importorskip("matplotlib")

        self.nybb = geopandas.read_file(geodatasets.get_path("nybb"))
        self.G = graph.Graph.build_contiguity(self.nybb)

        self.nybb_str = self.nybb.set_index("BoroName")
        self.G_str = graph.Graph.build_contiguity(self.nybb_str)

        self.expected_paths = [
            np.array(
                [
                    [943802.68511489, 147890.05410767],
                    [943802.68511489, 147890.05410767],
                ]
            ),
            np.array(
                [
                    [1033983.96582281, 196127.39050293],
                    [998506.94007314, 177674.69769287],
                ]
            ),
            np.array(
                [
                    [1033983.96582281, 196127.39050293],
                    [995258.50398262, 226631.05230713],
                ]
            ),
            np.array(
                [
                    [1033983.96582281, 196127.39050293],
                    [1021230.82521895, 251186.3362925],
                ]
            ),
            np.array(
                [
                    [998506.94007314, 177674.69769287],
                    [995258.50398262, 226631.05230713],
                ]
            ),
            np.array(
                [
                    [995258.50398262, 226631.05230713],
                    [1021230.82521895, 251186.3362925],
                ]
            ),
        ]

    def test_default(self):
        ax = self.G.plot(self.nybb)
        # nodes and edges
        assert len(ax.collections) == 2

        # edge geom
        linecollection = ax.collections[0]
        paths = linecollection.get_paths()
        for i, path in enumerate(paths):
            np.testing.assert_almost_equal(self.expected_paths[i], path.vertices)

        # node geom
        pathcollection = ax.collections[1]
        np.testing.assert_array_equal(
            pathcollection.get_offsets().data,
            shapely.get_coordinates(self.nybb.representative_point()),
        )

        # edge color
        np.testing.assert_array_equal(
            linecollection.get_color(), np.array([[0.0, 0.0, 0.0, 1.0]])
        )

        # node color
        np.testing.assert_array_equal(
            pathcollection.get_facecolor(), np.array([[0.0, 0.0, 0.0, 1.0]])
        )

    def test_string_id(self):
        ax = self.G_str.plot(self.nybb_str)
        assert len(ax.collections) == 2

        linecollection = ax.collections[0]
        paths = linecollection.get_paths()
        for i, path in enumerate(paths):
            np.testing.assert_almost_equal(self.expected_paths[i], path.vertices)

        pathcollection = ax.collections[1]
        np.testing.assert_array_equal(
            pathcollection.get_offsets().data,
            shapely.get_coordinates(self.nybb.representative_point()),
        )

    def test_misaligned(self):
        ax = self.G_str.plot(self.nybb_str.sort_index())
        assert len(ax.collections) == 2

        linecollection = ax.collections[0]
        paths = linecollection.get_paths()
        for i, path in enumerate(paths):
            np.testing.assert_almost_equal(self.expected_paths[i], path.vertices)

        pathcollection = ax.collections[1]
        np.testing.assert_array_equal(
            pathcollection.get_offsets().data,
            shapely.get_coordinates(self.nybb.representative_point()),
        )

    def test_no_nodes(self):
        ax = self.G.plot(self.nybb, nodes=False)
        assert len(ax.collections) == 1

        linecollection = ax.collections[0]
        paths = linecollection.get_paths()
        for i, path in enumerate(paths):
            np.testing.assert_almost_equal(self.expected_paths[i], path.vertices)

    def test_focal(self):
        ax = self.G_str.plot(self.nybb_str, focal="Queens")
        assert len(ax.collections) == 3

        linecollection = ax.collections[0]
        paths = linecollection.get_paths()
        np.testing.assert_almost_equal(self.expected_paths[1], paths[0].vertices)
        np.testing.assert_almost_equal(self.expected_paths[2], paths[1].vertices)
        np.testing.assert_almost_equal(self.expected_paths[3], paths[2].vertices)
        assert len(paths) == 3

        pathcollection = ax.collections[1]
        np.testing.assert_array_equal(
            pathcollection.get_offsets().data,
            shapely.get_coordinates(self.nybb.representative_point())[2:],
        )

        pathcollection_focal = ax.collections[2]
        np.testing.assert_array_equal(
            pathcollection_focal.get_offsets().data,
            shapely.get_coordinates(self.nybb.representative_point())[[1]],
        )

    def test_focal_array(self):
        ax = self.G_str.plot(self.nybb_str, focal=["Queens", "Bronx"])
        assert len(ax.collections) == 3

        linecollection = ax.collections[0]
        paths = linecollection.get_paths()
        np.testing.assert_almost_equal(self.expected_paths[1], paths[0].vertices)
        np.testing.assert_almost_equal(self.expected_paths[2], paths[1].vertices)
        np.testing.assert_almost_equal(self.expected_paths[3], paths[2].vertices)
        np.testing.assert_almost_equal(self.expected_paths[5], paths[3].vertices)
        assert len(paths) == 4

        pathcollection = ax.collections[1]
        np.testing.assert_array_equal(
            pathcollection.get_offsets().data,
            shapely.get_coordinates(self.nybb.representative_point())[1:],
        )

        pathcollection_focal = ax.collections[2]
        np.testing.assert_array_equal(
            pathcollection_focal.get_offsets().data,
            shapely.get_coordinates(self.nybb.representative_point())[[1, -1]],
        )

    def test_color(self):
        ax = self.G.plot(self.nybb, color="red")

        linecollection = ax.collections[0]
        np.testing.assert_array_equal(
            linecollection.get_color(), np.array([[1.0, 0.0, 0.0, 1.0]])
        )

        pathcollection = ax.collections[1]
        np.testing.assert_array_equal(
            pathcollection.get_facecolor(), np.array([[1.0, 0.0, 0.0, 1.0]])
        )

    def test_kws(self):
        ax = self.G.plot(
            self.nybb, edge_kws={"linestyle": "dotted"}, node_kws={"marker": "+"}
        )

        linecollection = ax.collections[0]
        assert linecollection.get_linestyle() == [(0.0, [1.5, 2.4749999999999996])]

        pathcollection = ax.collections[1]
        np.testing.assert_array_equal(
            pathcollection.get_paths()[0].vertices,
            np.array([[-0.5, 0.0], [0.5, 0.0], [0.0, -0.5], [0.0, 0.5]]),
        )

    def test_ax(self):
        ax = self.nybb.plot()
        self.G.plot(self.nybb, ax=ax)
        assert len(ax.collections) == 3

        # edge geom
        linecollection = ax.collections[1]
        paths = linecollection.get_paths()
        for i, path in enumerate(paths):
            np.testing.assert_almost_equal(self.expected_paths[i], path.vertices)

        # node geom
        pathcollection = ax.collections[2]
        np.testing.assert_array_equal(
            pathcollection.get_offsets().data,
            shapely.get_coordinates(self.nybb.representative_point()),
        )

    def test_figsize(self):
        ax = self.G.plot(self.nybb, figsize=(12, 12))
        np.testing.assert_array_equal(
            ax.figure.get_size_inches(), np.array([12.0, 12.0])
        )

    def test_limit_extent(self):
        ax = self.G_str.plot(self.nybb_str)
        self.G_str.plot(
            self.nybb_str, focal="Bronx", limit_extent=True, ax=ax, color="red"
        )
        assert ax.get_ylim() == (193374.44321345096, 253939.28358198277)
        assert ax.get_xlim() == (993322.2308906089, 1035920.2389148154)

    def test_focal_kws(self):
        ax = self.G_str.plot(
            self.nybb_str,
            focal="Queens",
            focal_kws={"color": "blue"},
            node_kws={"edgecolor": "pink"},
        )

        pathcollection = ax.collections[1]
        np.testing.assert_array_almost_equal(
            pathcollection.get_edgecolor(),
            np.array([[1.0, 0.75294118, 0.79607843, 1.0]]),
        )

        pathcollection_focal = ax.collections[2]
        # inherit node_kws
        np.testing.assert_array_almost_equal(
            pathcollection_focal.get_edgecolor(),
            np.array([[1.0, 0.75294118, 0.79607843, 1.0]]),
        )
        # apply own kws
        np.testing.assert_array_equal(
            pathcollection_focal.get_facecolor(), np.array([[0.0, 0.0, 1.0, 1.0]])
        )


class TestExplore:
    def setup_method(self):
        # skip tests when no folium installed
        pytest.importorskip("folium")

        self.nybb_str = geopandas.read_file(geodatasets.get_path("nybb")).set_index(
            "BoroName"
        )
        self.G_str = graph.Graph.build_contiguity(self.nybb_str)

    def test_default(self):
        m = self.G_str.explore(self.nybb_str)
        s = fetch_map_string(m)

        # nodes
        assert s.count("Point") == 5
        # edges
        assert s.count("LineString") == 6
        # tooltip
        assert '"focal":"Queens","neighbor":"Bronx","weight":1}' in s
        # color
        assert s.count('"__folium_color":"black"') == 11
        # labels
        assert s.count("Brooklyn") == 3

    def test_no_nodes(self):
        m = self.G_str.explore(self.nybb_str, nodes=False)
        s = fetch_map_string(m)

        # nodes
        assert s.count("Point") == 0
        # edges
        assert s.count("LineString") == 6
        # tooltip
        assert '"focal":"Queens","neighbor":"Bronx","weight":1}' in s
        # color
        assert s.count('"__folium_color":"black"') == 6
        # labels
        assert s.count("Brooklyn") == 2

    def test_focal(self):
        m = self.G_str.explore(self.nybb_str, focal="Queens")
        s = fetch_map_string(m)

        # nodes
        assert s.count("Point") == 4
        # edges
        assert s.count("LineString") == 3
        # tooltip
        assert '"focal":"Queens","neighbor":"Bronx","weight":1}' in s
        assert '"focal":"Queens","neighbor":"Manhattan","weight":1}' in s
        assert '"focal":"Queens","neighbor":"Brooklyn","weight":1}' in s
        # color
        assert s.count('"__folium_color":"black"') == 7
        # labels
        assert s.count("Brooklyn") == 2

    def test_focal_array(self):
        m = self.G_str.explore(self.nybb_str, focal=["Queens", "Bronx"])
        s = fetch_map_string(m)

        # if node is both focal and neighbor, both are plottted as you can style
        # them differently to see both
        assert s.count("Point") == 6
        # edges
        assert s.count("LineString") == 4
        # tooltip
        assert '"focal":"Queens","neighbor":"Bronx","weight":1}' in s
        assert '"focal":"Queens","neighbor":"Manhattan","weight":1}' in s
        assert '"focal":"Queens","neighbor":"Brooklyn","weight":1}' in s
        assert '"focal":"Bronx","neighbor":"Manhattan","weight":1}' in s

        # color
        assert s.count('"__folium_color":"black"') == 10
        # labels
        assert s.count("Brooklyn") == 2

    def test_color(self):
        m = self.G_str.explore(self.nybb_str, color="red")
        s = fetch_map_string(m)

        assert s.count('"__folium_color":"red"') == 11

    def test_kws(self):
        m = self.G_str.explore(
            self.nybb_str,
            focal=["Queens", "Bronx"],
            edge_kws={"color": "red"},
            node_kws={"color": "blue", "marker_kwds": {"radius": 8}},
            focal_kws={"color": "pink", "marker_kwds": {"radius": 12}},
        )
        s = fetch_map_string(m)

        # color
        assert s.count('"__folium_color":"red"') == 4
        assert s.count('"__folium_color":"blue"') == 4
        assert s.count('"__folium_color":"pink"') == 2

        assert '"radius":8' in s
        assert '"radius":12' in s

    def test_m(self):
        m = self.nybb_str.explore()
        self.G_str.explore(self.nybb_str, m=m)
        s = fetch_map_string(m)

        # nodes
        assert s.count("Point") == 5
        # edges
        assert s.count("LineString") == 6
        # geoms
        assert s.count("Polygon") == 5

    def test_explore_kwargs(self):
        m = self.G_str.explore(self.nybb_str, tiles="OpenStreetMap HOT")
        s = fetch_map_string(m)

        assert "tile.openstreetmap.fr/hot" in s



"""Unit test for raster.py"""

import numpy as np
import pandas as pd
import pytest

from .. import raster


class Testraster:
    def setup_method(self):
        pytest.importorskip("xarray")
        self.da1 = raster.testDataArray()
        self.da2 = raster.testDataArray((1, 4, 4), missing_vals=False)
        self.da3 = self.da2.rename({"band": "layer", "x": "longitude", "y": "latitude"})
        self.data1 = pd.Series(np.ones(5))
        self.da4 = raster.testDataArray((1, 1), missing_vals=False)
        self.da4.data = np.array([["test"]])

    def test_da2_w(self):
        w1 = raster.da2W(self.da1, "queen", k=2, n_jobs=-1)
        assert w1[(1, -30.0, -180.0)] == {(1, -90.0, 60.0): 1, (1, -90.0, -60.0): 1}
        assert w1[(1, -30.0, 180.0)] == {(1, -90.0, -60.0): 1, (1, -90.0, 60.0): 1}
        assert w1.n == 5
        assert w1.index.names == self.da1.to_series().index.names
        assert w1.index.tolist()[0] == (1, 90.0, 180.0)
        assert w1.index.tolist()[1] == (1, -30.0, -180.0)
        assert w1.index.tolist()[2] == (1, -30.0, 180.0)
        assert w1.index.tolist()[3] == (1, -90.0, -60.0)
        w2 = raster.da2W(self.da2, "rook")
        assert sorted(w2.neighbors[(1, -90.0, 180.0)]) == [
            (1, -90.0, 60.0),
            (1, -30.0, 180.0),
        ]
        assert sorted(w2.neighbors[(1, -90.0, 60.0)]) == [
            (1, -90.0, -60.0),
            (1, -90.0, 180.0),
            (1, -30.0, 60.0),
        ]
        assert w2.n == 16
        assert w2.index.names == self.da2.to_series().index.names
        assert w2.index.tolist() == self.da2.to_series().index.tolist()
        coords_labels = {
            "z_label": "layer",
            "y_label": "latitude",
            "x_label": "longitude",
        }
        w3 = raster.da2W(self.da3, z_value=1, coords_labels=coords_labels)
        assert sorted(w3.neighbors[(1, -90.0, 180.0)]) == [
            (1, -90.0, 60.0),
            (1, -30.0, 60.0),
            (1, -30.0, 180.0),
        ]
        assert w3.n == 16
        assert w3.index.names == self.da3.to_series().index.names
        assert w3.index.tolist() == self.da3.to_series().index.tolist()

    def test_da2_wsp(self):
        w1 = raster.da2WSP(self.da1, "rook", n_jobs=-1)
        rows, cols = w1.sparse.shape
        n = rows * cols
        pct_nonzero = w1.sparse.nnz / float(n)
        assert pct_nonzero == 0.08
        data = w1.sparse.todense().tolist()
        assert data[3] == [0, 0, 0, 0, 1]
        assert data[4] == [0, 0, 0, 1, 0]
        assert w1.index.names == self.da1.to_series().index.names
        assert w1.index.tolist()[0] == (1, 90.0, 180.0)
        assert w1.index.tolist()[1] == (1, -30.0, -180.0)
        assert w1.index.tolist()[2] == (1, -30.0, 180.0)
        assert w1.index.tolist()[3] == (1, -90.0, -60.0)
        w2 = raster.da2WSP(self.da2, "queen", k=2, include_nodata=True)
        w3 = raster.da2WSP(self.da2, "queen", k=2, n_jobs=-1)
        assert w2.sparse.nnz == w3.sparse.nnz
        assert w2.sparse.todense().tolist() == w3.sparse.todense().tolist()
        assert w2.n == 16
        assert w2.index.names == self.da2.to_series().index.names
        assert w2.index.tolist() == self.da2.to_series().index.tolist()

    def test_w2da(self):
        xarray = pytest.importorskip("xarray")
        w2 = raster.da2W(self.da2, "rook", n_jobs=-1)
        da2 = raster.w2da(self.da2.data.flatten(), w2, self.da2.attrs, self.da2.coords)
        da_compare = xarray.DataArray.equals(da2, self.da2)
        assert da_compare is True

    def test_wsp2da(self):
        wsp1 = raster.da2WSP(self.da1, "queen")
        da1 = raster.wsp2da(self.data1, wsp1)
        assert da1["y"].values.tolist() == self.da1["y"].values.tolist()
        assert da1["x"].values.tolist() == self.da1["x"].values.tolist()
        assert da1.shape == (1, 4, 4)

    def test_da_checker(self):
        pytest.raises(ValueError, raster.da2W, self.da4)
